__all__ = ()

from ..place_holder import PlaceHolder

from .preinstance import Preinstance


def _get_direct_type_parent(meta_type, type_parents):
    """
    Gets the direct parent from the given parents.
    
    Parameters
    ----------
    meta_type : `type`
        The type that is being instantiated.
    
    type_parents : `tuple<type>`
        The type's parents.
    
    Returns
    -------
    type_parent : `None | type`
    """
    if type_parents:
        type_parent = type_parents[0]
        if not isinstance(type_parent, meta_type):
            type_parent = None
    else:
        type_parent = None
    
    return type_parent


def _pop_items_to_post_instantiate(type_attributes):
    """
    Pops the items from the type's attributes that should be instantiated after the type is created.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        The defined attributes of the type.
    
    Returns
    -------
    to_post_instantiate : `list<(str, Preinstance)>`
    """
    to_post_instantiate = []
    
    for attribute_name, attribute_value in type_attributes.items():
        if isinstance(attribute_value, Preinstance):
            to_post_instantiate.append((attribute_name, attribute_value))
    
    for item in to_post_instantiate:
        del type_attributes[item[0]]
    
    return to_post_instantiate



def _get_name_default(type_parent, name_default):
    """
    Gets the default name to use for post created instances.
    
    Parameters
    ----------
    type_parent : `None | type`
        Parent type.
    
    name_default : `str | ...`
        The passed name default.
    
    Returns
    -------
    name_default : `str`
    
    Raises
    ------
    TypeError
    """
    if (name_default is not ...):
        if (type(name_default) is not str):
            raise TypeError(f'`name_default` can be only `str`, got {type(name_default).__name__}')
    
    elif (type_parent is not None):
        name_default = type_parent.NAME_DEFAULT
    
    else:
        name_default = 'UNDEFINED'
    
    return name_default


def _identify_value_type(type_parent, base_type, value_type):
    """
    Identifies the value type.
    
    Parameters
    ----------
    type_parent : `None | type`
        Parent type.
    
    base_type : `bool`
        Whether the currently created type is a base type.
    
    value_type : `type | ...`
        The passed value type.
    
    Returns
    -------
    value_type : `type<NoneType | int | str>`
    
    Raises
    ------
    RuntimeError
    TypeError
    """
    if (
        (value_type is ...) and
        (not base_type) and
        (
            (type_parent is None) or
            (type_parent.VALUE_TYPE is type(None))
        )
    ):
        raise RuntimeError(
            '`value_type` must be given if `base_type = False` (so by default) '
            'and if non of its parents defined it.',
        )
    
    while True:
        if value_type is ...:
            if type_parent is None:
                value_type = type(None)
            else:
                value_type = type_parent.VALUE_TYPE
            break
        
        if value_type is int:
            break
        
        if value_type is str:
            break
        
        raise TypeError(
            f'`value_type` not supported. Can be `int` or `str`; got {value_type.__name__}.'
        )
    
    
    if (
        (value_type is not ...) and
        (type_parent is not None) and
        (type_parent.VALUE_TYPE is not type(None)) and
        (value_type is not type_parent.VALUE_TYPE)
    ):
        raise TypeError(
            f'`value_type` cannot diverge from parent\'s; '
            f'got value_type = {value_type.__name__}; type_parent.value_type = {type_parent.VALUE_TYPE.__name__}.'
        )
    
    return value_type


def _get_value_default(value_type):
    """
    Gets the default value of the given type.
    
    Parameters
    ----------
    value_type : `type<NoneType | int | str>`
        The value's type.
    
    Returns
    -------
    value_default : `None | int | str`
    """
    if value_type is type(None):
        value_default = None
    
    elif value_type is int:
        value_default = 0
    
    elif value_type is str:
        value_default = ''
    
    else:
        value_default = None
    
    return value_default


def _get_and_validate_slots(type_attributes):
    """
    Gets the slots from the type attributes and validates them.
    If missing inserts empty slots.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    Returns
    -------
    slots : `tuple<str>`
    
    Raises
    ------
    TypeError
    """
    try:
        slots = type_attributes['__slots__']
    except KeyError:
        slots = ()
        type_attributes['__slots__'] = slots
    else:
        if not isinstance(slots, tuple):
            raise TypeError(
                f'Slots can be `tuple<str>`; got {type(slots).__name__}; slots = {slots!r}.'
            )
        
        for element in slots:
            if type(element) is not str:
                raise TypeError(
                    f'Slots elements can be `str`; got {type(element).__name__}; slots = {slots!r}.'
                )
    
    return slots


def _set_slot_place_holders(type_attributes, type_parent, slots, name_default, value_default):
    """
    Sets slot placeholders to type attributes if required.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    type_parent : `None | type`
        Parent type.
    
    slots : `tuple<str>`
        The defined slots of the type.
    
    name_default : `str`
        The default name to use.
    
    value_default : `int | str`
        The default value to use.
    """
    if (type_parent is not None):
        return
    
    if 'name' not in slots:
        type_attributes['name'] = PlaceHolder(name_default)
    
    if 'value' not in slots:
        type_attributes['value'] = PlaceHolder(value_default)


def _inherit_hash_function(type_attributes, type_parent):
    """
    Inherits the parent's hash method of the parent type if required.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    type_parent : `None | type`
        Parent type.
    """
    if (type_parent is None) or ('__hash__' in type_attributes.keys()):
        return
    
    type_attributes['__hash__'] = type_parent.__hash__


class PreinstancedMeta(type):
    """
    Meta-type for preinstanced instances.
    """
    def __new__(
        cls,
        type_name,
        type_parents,
        type_attributes,
        *,
        base_type = False,
        name_default = ...,
        value_type = ...,
    ):
        """
        Creates a preinstanced type.
        
        Parameters
        ----------
        type_name : `str`
            The created type's name.
        
        type_parents : `tuple` of `type`
            The parents of the creates type.
        
        type_attributes : `dict<str, object>`
            The type attributes of the created type.
        
        base_type : `bool` = `False`, Optional (Keyword only)
            Whether a base type is being created.
        
        name_default : `str`, Optional (Keyword only)
            Default name to use for preinstanced objects.
        
        value_type : `int | str`, Optional (Keyword only)
            The instance's value's type.
        
        Returns
        -------
        type : `instance<cls>`
        """
        type_parent = _get_direct_type_parent(cls, type_parents)
        
        if (type_parent is not None) and (type_parent.INSTANCES is not NotImplemented):
            raise RuntimeError(
                f'Cannot inherit from a non-base {PreinstancedMeta.__name__} type.'
            )
        
        name_default = _get_name_default(type_parent, name_default)
        value_type = _identify_value_type(type_parent, base_type, value_type)
        value_default = _get_value_default(value_type)
        
        to_post_instantiate = _pop_items_to_post_instantiate(type_attributes)
        
        # add attributes
        type_attributes['INSTANCES'] = NotImplemented if base_type else {}
        type_attributes['VALUE_TYPE'] = type(None) if (value_type is ...) else value_type
        type_attributes['NAME_DEFAULT'] = name_default
        type_attributes['VALUE_DEFAULT'] = value_default
        
        slots = _get_and_validate_slots(type_attributes)
        _set_slot_place_holders(type_attributes, type_parent, slots, name_default, value_default)
        _inherit_hash_function(type_attributes, type_parent)
        
        type_ = type.__new__(cls, type_name, type_parents, type_attributes)
        
        # add instances
        for attribute_name, attribute_value in to_post_instantiate:
            setattr(
                type_,
                attribute_name,
                type_(
                    attribute_value.value,
                    attribute_value.name,
                    *attribute_value.positional_parameters,
                    **attribute_value.keyword_parameters,
                ),
            )
        
        return type_
    
    
    def __call__(cls, value = None, name = None, *positional_parameters, **keyword_parameters):
        """
        Creates a new preinstanced instance if allowed and stores it.
        
        Parameters
        ----------
        value : `None | .VALUE_TYPE` = `None`, Optional
            The value of the preinstanced object.
        
        name : `None | str` = `None`, Optional
            The object's name.
        
        *positional_parameters : Positional parameters
            Additional positional parameters to create the instance with.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the instance with.
        
        Raises
        ------
        NotImplementedError
        TypeError
        """
        INSTANCES = cls.INSTANCES
        if INSTANCES is NotImplemented:
            raise NotImplementedError(
                f'Base types cannot be instantiated; type = {cls.__name__}.'
            )
        
        if value is None:
            value = cls.VALUE_DEFAULT
        
        try:
            return INSTANCES[value]
        except KeyError:
            pass
        
        new = cls.__new__(cls, value, name, *positional_parameters, **keyword_parameters)
        INSTANCES[value] = new
        return new



__all__ = ()

from scarletio import include

from ..emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data
from ..field_parsers import (
    nullable_functional_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory
)
from ..field_putters import (
    nullable_entity_array_putter_factory, nullable_functional_optional_putter_factory, url_optional_putter_factory
)
from ..field_validators import (
    nullable_entity_validator_factory, nullable_object_array_validator_factory, nullable_string_validator_factory
)

from .shared_constants import CUSTOM_ID_LENGTH_MAX


Component = include('Component')


# components

parse_components = nullable_object_array_parser_factory('components', NotImplemented, include = 'Component')
put_components_into = nullable_entity_array_putter_factory('components', NotImplemented, include = 'Component')


def validate_components(components):
    """
    Validates the given components.
    
    Parameters
    ----------
    components : `None | iterable` of `instance<Component>`
        The components to validate.
    
    Returns
    -------
    components_processed : `None | tuple<Component>`
    
    Raises
    ------
    TypeError
        - If `components` is invalid type.
    ValueError
        - If `components` contains an invalid value.
    """
    if components is None:
        return None
    
    if (getattr(components, '__iter__', None) is None):
        raise TypeError(
            f'`components` can be `None`, `iterable` of `{Component.__name__}`, got '
            f'{type(components).__name__}; {components!r}.'
        )
        
    components_processed = None
    
    for component in components:
        if not isinstance(component, Component):
            raise TypeError(
                f'`components` can contain `{Component.__name__}` elements, got '
                f'{type(component).__name__}; {component!r}; components = {components!r}.'
            )
        
        if not component.type.layout_flags.nestable:
            raise ValueError(
                f'Cannot nest components of type {component.type.name}, '
                f'got {component!r}; components = {components!r}.'
            )
        
        if (components_processed is None):
            components_processed = []
        
        components_processed.append(component)
    
    if (components_processed is not None):
        components_processed = tuple(components_processed)
    
    return components_processed


# custom_id

parse_custom_id = nullable_string_parser_factory('custom_id')
put_custom_id_into = url_optional_putter_factory('custom_id')
validate_custom_id = nullable_string_validator_factory('custom_id', 0, CUSTOM_ID_LENGTH_MAX)

# emoji

parse_emoji = nullable_functional_parser_factory('emoji', create_partial_emoji_from_data)
put_emoji_into = nullable_functional_optional_putter_factory('emoji', create_partial_emoji_data)
validate_emoji = nullable_entity_validator_factory('emoji', Emoji)



__all__ = ('OperationSystem',)

from ...bases import Preinstance as P, PreinstancedBase


class OperationSystem(PreinstancedBase, value_type = str):
    """
    Represents a ``ApplicationExecutable``'s operation system.
    
    Attributes
    ----------
    name : `str`
        The name of the operation system.
    
    value : `str`
        The Discord side identifier value of the os.
    
    Type Attributes
    ---------------
    Every predefined os can be accessed as type attribute as well:
    +-----------------------+-----------+---------------+
    | Type attribute name   | name      | value         |
    +=======================+===========+===============+
    | none                  | none      | `''`          |
    +-----------------------+-----------+---------------+
    | darwin                | darwin    | '`darwin'`    |
    +-----------------------+-----------+---------------+
    | linux                 | linux     | '`linux`'     |
    +-----------------------+-----------+---------------+
    | windows               | windows   | `'win32'`     |
    +-----------------------+-----------+---------------+
    """
    __slots__ = ()
    
    # predefined
    none = P('', 'none')
    darwin = P('darwin', 'darwin')
    linux = P('linux', 'linux')
    windows = P('win32', 'windows')



__all__ = ()

from ...emoji import Emoji
from ...field_parsers import (
    entity_id_parser_factory, force_string_parser_factory, int_parser_factory, nullable_string_parser_factory,
    preinstanced_array_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, force_string_putter_factory, int_putter_factory, nullable_string_putter_factory,
    preinstanced_array_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, int_conditional_validator_factory,
    nullable_string_validator_factory, preinstanced_array_validator_factory
)
from ...sticker import Sticker

from .constants import DESCRIPTION_LENGTH_MAX, NAME_LENGTH_MAX

# approximate_online_count

parse_approximate_online_count = int_parser_factory('approximate_presence_count', 0)
put_approximate_online_count = int_putter_factory('approximate_presence_count')
validate_approximate_online_count = int_conditional_validator_factory(
    'approximate_online_count',
    0,
    lambda approximate_online_count : approximate_online_count >= 0,
    '>= 0',
)

# approximate_user_count

parse_approximate_user_count = int_parser_factory('approximate_member_count', 0)
put_approximate_user_count = int_putter_factory('approximate_member_count')
validate_approximate_user_count = int_conditional_validator_factory(
    'approximate_user_count',
    0,
    lambda approximate_user_count : approximate_user_count >= 0,
    '>= 0',
)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# emojis

def parse_emojis(data, entities, guild_id):
    """
    Parses the `emojis` field out from the given guild preview data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Guild preview data.
    entities : `dict` of (`int`, ``Emoji``) items
        The entity container to populate.
    guild_id : `int`
        The respective guild's identifier.
    
    Returns
    -------
    entities : `dict` of (`int`, ``Emoji``) items
    """
    emoji_datas = data.get('emojis', None)
    if emoji_datas is None:
        return entities
    
    if entities:
        entity_cache = [*entities.values()]
        entities.clear()
    
    for emoji_data in emoji_datas:
        emoji = Emoji.from_data(emoji_data, guild_id)
        entities[emoji.id] = emoji
    
    return entities


def put_emojis(entities, data, defaults):
    """
    Puts the given emojis to the guild preview data.
    
    Parameters
    ----------
    entities : `dict` of (`int`, ``Emoji``) items
        Emojis to put into the given data.
    data : `dict` of (`str`, `object`) items
        Guild preview data.
    defaults : `bool`
        Whether values with their default fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['emojis'] = [
        entity.to_data(defaults = defaults, include_internals = True) for entity in sorted(entities.values())
    ]
    return data


def validate_emojis(emojis):
    """
    Validates the given emojis.
    
    Parameters
    ----------
    emojis : `None`, `iterable` of ``Emoji``
        The emojis to validate.
    
    Returns
    -------
    emojis : `dict` of (`int`, ``Emoji``) items
        The validated emojis.
    
    Raises
    ------
    TypeError
        - If `emojis` is not `None`, `iterable` of ``Emoji``.
        - If `emojis` contains a non-``Emoji`` element. 
    """
    validated_emojis = {}
    
    if emojis is None:
        return validated_emojis
    
    if getattr(type(emojis), '__iter__', None) is None:
        raise TypeError(
            f'`emojis` can be `None` ot `iterable` of `{Emoji.__name__}`, '
            f'got {emojis.__class__.__name__}; {emojis!r}.'
        )
    
    for emoji in emojis:
        if not isinstance(emoji, Emoji):
            raise TypeError(
                f'`emoji` can contain `{Emoji.__name__}` elements, got {emoji.__class__.__name__}; {emoji!r}; '
                f'emojis = {emojis!r}.'
            )
        
        validated_emojis[emoji.id] = emoji
    
    return validated_emojis

# features

parse_features = preinstanced_array_parser_factory('features', NotImplemented, include = 'GuildFeature')
put_features = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', NotImplemented, include = 'GuildFeature')

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id', NotImplemented, include = 'Guild')

# stickers

def parse_stickers(data, entities):
    """
    Parses the `stickers` field out from the given guild preview data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Guild preview data.
    entities : `dict` of (`int`, ``Sticker``) items
        The entity container to populate.
    
    Returns
    -------
    entities : `dict` of (`int`, ``Sticker``) items
    """
    sticker_datas = data.get('stickers', None)
    if sticker_datas is None:
        return entities
    
    if entities:
        entity_cache = [*entities.values()]
        entities.clear()
    
    for sticker_data in sticker_datas:
        sticker = Sticker.from_data(sticker_data)
        entities[sticker.id] = sticker
    
    return entities


def put_stickers(entities, data, defaults):
    """
    Puts the given stickers to the guild preview data.
    
    Parameters
    ----------
    entities : `dict` of (`int`, ``Sticker``) items
        Stickers to put into the given data.
    data : `dict` of (`str`, `object`) items
        Guild preview data.
    defaults : `bool`
        Whether values with their default fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['stickers'] = [
        entity.to_data(defaults = defaults, include_internals = True) for entity in sorted(entities.values())
    ]
    return data


def validate_stickers(stickers):
    """
    Validates the given stickers.
    
    Parameters
    ----------
    stickers : `None`, `iterable` of ``Sticker``
        The stickers to validate.
    
    Returns
    -------
    stickers : `dict` of (`int`, ``Sticker``) items
        The validated stickers.
    
    Raises
    ------
    TypeError
        - If `stickers` is not `None`, `iterable` of ``Sticker``.
        - If `stickers` contains a non-``Sticker`` element. 
    """
    validated_stickers = {}
    
    if stickers is None:
        return validated_stickers
    
    if getattr(type(stickers), '__iter__', None) is None:
        raise TypeError(
            f'`stickers` can be `None` ot `iterable` of `{Sticker.__name__}`, '
            f'got {type(stickers).__name__}; {stickers!r}.'
        )
    
    if isinstance(stickers, dict):
        iterator = iter(stickers.values())
    else:
        iterator = iter(stickers)
    
    for sticker in iterator:
        if not isinstance(sticker, Sticker):
            raise TypeError(
                f'`sticker` can contain `{Sticker.__name__}` elements, got {type(sticker).__name__}; {sticker!r}; '
                f'stickers = {stickers!r}.'
            )
        
        validated_stickers[sticker.id] = sticker
    
    return validated_stickers


# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', 0, NAME_LENGTH_MAX)



__all__ = ('AutoModerationEventType', 'AutoModerationRuleTriggerType')

from ...bases import Preinstance as P, PreinstancedBase

from ..trigger_metadata import (
    AutoModerationRuleTriggerMetadataBase, AutoModerationRuleTriggerMetadataKeywordPreset,
    AutoModerationRuleTriggerMetadataKeyword, AutoModerationRuleTriggerMetadataMentionSpam, 
)


class AutoModerationRuleTriggerType(PreinstancedBase, value_type = int):
    """
    Represents an auto moderation rule's trigger type.
    
    Attributes
    ----------
    max_per_guild : `int`
        The maximal amount of rules of this type per guild.
    
    metadata_type : `type<AutoModerationRuleTriggerMetadataBase>`
        The trigger type's respective metadata type.
    
    name : `str`
        The default name of the auto moderation trigger type.
    
    value : `int`
        The Discord side identifier value of the auto moderation trigger type.
    
    Type Attributes
    ---------------
    Every predefined auto moderation trigger type is also stored as a type attribute:
    
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | Type attribute name   | Name              | Value     | Max per guild | Metadata type                                         |
    +=======================+===================+===========+===============+=======================================================+
    | none                  | none              | 0         | 0             | ``AutoModerationRuleTriggerMetadataBase``             |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | keyword               | keyword           | 1         | 5             | ``AutoModerationRuleTriggerMetadataKeyword``          |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | harmful_link          | harmful link      | 2         | 1             | ``AutoModerationRuleTriggerMetadataBase``             |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | spam                  | spam              | 3         | 1             | ``AutoModerationRuleTriggerMetadataBase``             |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | keyword_preset        | keyword preset    | 4         | 1             | ``AutoModerationRuleTriggerMetadataKeywordPreset``    |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | mention_spam          | mention spam      | 5         | 1             | ``AutoModerationRuleTriggerMetadataMentionSpam``      |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    | user_profile          | user profile      | 6         | 1             | ``AutoModerationRuleTriggerMetadataKeyword``          |
    +-----------------------+-------------------+-----------+---------------+-------------------------------------------------------+
    """
    __slots__ = ('max_per_guild', 'metadata_type')
    
    def __new__(cls, value, name = None, max_per_guild = 1, metadata_type = None):
        """
        Creates an auto moderation rule trigger type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the auto moderation trigger type.
        
        name : `None | str` = `None`, Optional
            The default name of the auto moderation trigger type.
        
        max_per_guild : `int` = `1`, Optional
            The native name of the auto moderation trigger type.
        
        metadata_type : `None | type<AutoModerationRuleTriggerMetadataBase>` = `None`, Optional
            The trigger type's respective metadata type.
        """
        if metadata_type is None:
            metadata_type = AutoModerationRuleTriggerMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.max_per_guild = max_per_guild
        self.metadata_type = metadata_type
        return self
    
    
    # predefined
    none = P(0, 'none', 0, AutoModerationRuleTriggerMetadataBase)
    keyword = P(1, 'keyword', 5, AutoModerationRuleTriggerMetadataKeyword)
    harmful_link = P(2, 'harmful link', 1, AutoModerationRuleTriggerMetadataBase)
    spam = P(3, 'spam', 1, AutoModerationRuleTriggerMetadataBase)
    keyword_preset = P(4, 'keyword preset', 1, AutoModerationRuleTriggerMetadataKeywordPreset)
    mention_spam = P(5, 'mention spam', 1, AutoModerationRuleTriggerMetadataMentionSpam)
    user_profile = P(6, 'user profile', 1, AutoModerationRuleTriggerMetadataKeyword)


class AutoModerationEventType(PreinstancedBase, value_type = int):
    """
    Represents an auto moderation rule's event type.
    
    Attributes
    ----------
    name : `str`
        The default name of the auto moderation event type.
    
    value : `int`
        The Discord side identifier value of the auto moderation event type.
    
    Type Attributes
    ---------------
    Every predefined auto moderation event type is also stored as a type attribute:
    
    +-----------------------+-------------------+-----------+
    | Type attribute name   | Name              | Value     |
    +=======================+===================+===========+
    | none                  | none              | 0         |
    +-----------------------+-------------------+-----------+
    | message_send          | message send      | 1         |
    +-----------------------+-------------------+-----------+
    | user_update           | user update       | 2         |
    +-----------------------+-------------------+-----------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    message_send = P(1, 'message send')
    user_update = P(2, 'user update')



from .component import *
from .component_metadata import *
from .entity_select_default_value import *
from .interaction_form import *
from .media_item import *
from .string_select_option import *

from .shared_constants import *
from .shared_fields import *
from .shared_helpers import *
from .utils import *


__all__ = (
    *component.__all__,
    *component_metadata.__all__,
    *entity_select_default_value.__all__,
    *interaction_form.__all__,
    *media_item.__all__,
    *string_select_option.__all__,
    
    *shared_constants.__all__,
    *shared_fields.__all__,
    *shared_helpers.__all__,
    *utils.__all__,
)



__all__ = ()

from .fields import put_emoji_pack_collectible, validate_emoji_pack_collectible


GUILD_INVENTORY_SETTINGS_FIELD_CONVERTERS = {
    'emoji_pack_collectible': (validate_emoji_pack_collectible, put_emoji_pack_collectible),
}



from .base import *
from .fields import *
from .location import *
from .stage import *
from .utils import *


__all__ = (
    *base.__all__,
    *fields.__all__,
    *location.__all__,
    *stage.__all__,
    *utils.__all__,
)



__all__ = ('EmbeddedActivityLocationType', )

from scarletio import copy_docs

from ...bases import Preinstance as P, PreinstancedBase


class EmbeddedActivityLocationType(PreinstancedBase, value_type = str):
    """
    Represents an embedded activity location's type.
    
    Attributes
    ----------
    name : `str`
        The embedded activity location type's name.
    
    value : `str`
        The Discord side identifier value of the embedded activity location type.
    
    Type Attributes
    ---------------
    Every predefined embedded activity location type can be accessed as type attribute as well:
    
    +-----------------------+-------------------+---------------+
    | Type attribute name   | name              | value         |
    +=======================+===================+===============+
    | none                  | none              |               |
    +-----------------------+-------------------+---------------+
    | guild_channel         | guild channel     | gc            |
    +-----------------------+-------------------+---------------+
    | private_channel       | private channel   | pc            |
    +-----------------------+-------------------+---------------+
    """
    __slots__ = ()
    
    @copy_docs(PreinstancedBase.__new__)
    def __new__(cls, value, name = None):
        if name is None:
            name = value.casefold().replace('-', ' ')
        
        return PreinstancedBase.__new__(cls, value, name)
    
    
    # predefined
    none = P('', 'none')
    guild_channel = P('gc', 'guild channel')
    private_channel = P('pc', 'private channel')



__all__ = ()

from scarletio import copy_docs

from .builder_base import BuilderBase

from .conversions import CONVERSION_KEYWORD, CONVERSION_NONE, CONVERSION_POSITIONAL


class BuilderFielded(BuilderBase):
    """
    Builder using fields to store its values.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        Fields already added to the builder.
    """
    __conversions_default__ = [
        CONVERSION_NONE,
        CONVERSION_KEYWORD,
        CONVERSION_POSITIONAL,
    ]
    
    __slots__ = ('fields',)
    
    def __new__(cls):
        """
        Creates a new message builder.
        """
        self = object.__new__(cls)
        self.fields = {}
        return self
    
    
    @copy_docs(BuilderBase._store_field_value)
    def _store_field_value(self, conversion, value):
        self.fields[conversion] = value
    
    
    @copy_docs(BuilderBase._try_pull_field_value)
    def _try_pull_field_value(self, conversion):
        try:
            yield self.fields[conversion]
        except KeyError:
            pass
    
    
    @copy_docs(BuilderBase._with_positional_parameter_unknown)
    def _with_positional_parameter_unknown(self, value):
        raise TypeError(
            f'Could not recognise value as a field. Got {type(value).__name__}; {value!r}.'
        )
        return self._setter_field(CONVERSION_CONTENT, str(value))
    
    
    @copy_docs(BuilderBase._with_keyword_parameter_unknown)
    def _with_keyword_parameter_unknown(self, key, value):
        raise TypeError(
            f'{key!r} could not be recognised as a field. Got {type(value).__name__}; {value!r}.'
        )
    
    
    @copy_docs(BuilderBase._iter_conversions)
    def _iter_conversions(self):
        yield from self.fields.keys()
    
    
    @copy_docs(BuilderBase._iter_fields)
    def _iter_fields(self):
        yield from self.fields.items()



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .preinstanced import MessageReferenceType


class MessageReferenceConfiguration(RichAttributeErrorBaseType):
    """
    Message reference configuration for message create endpoints.
    
    Attributes
    ----------
    channel_id : `int`
        The referenced message's channel's identifier. Used when forwarding.
    fail_fallback : `bool`
        Whether normal message should be sent if the referenced message is deleted. Used when replying.
    message_id : `int`
        The referenced message's identifier.
    message_reference_type : ``MessageReferenceType``
        What kind of reference does the instance represents.
    """
    __slots__ = ('channel_id', 'fail_fallback', 'message_id', 'type')
    
    def __new__(
        cls,
        *,
        channel_id = 0,
        fail_fallback = False,
        message_id = 0,
        message_reference_type = MessageReferenceType.reply,
    ):
        """
        Creates a new message reference.
        
        Parameters
        ----------
        channel_id : `int` = `0`, Optional (Keyword only)
            The referenced message's channel's identifier.
        fail_fallback : `bool` = `False`, Optional (Keyword only)
            Whether normal message should be sent if the referenced message is deleted.
        message_id : `int` = `0`, Optional (Keyword only)
            The referenced message's identifier.
        message_reference_type : ``MessageReferenceType`` = `MessageReferenceType.reply`, Optional (Keyword only)
            What kind of reference does the instance represents.
        """    
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.fail_fallback = fail_fallback
        self.message_id = message_id
        self.type = message_reference_type
        return self
        
    
    def to_data(self):
        """
        Serialises the message reference configuration.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {'message_id' : str(self.message_id)}
        
        message_reference_type = self.type
        
        if message_reference_type is MessageReferenceType.reply:
            if self.fail_fallback:
                data['fail_if_not_exists'] = False
        
        else:
            data['type'] = message_reference_type.value
            
            if message_reference_type is MessageReferenceType.forward:
                data['channel_id'] = str(self.channel_id)
        
        return data
    
    
    def __bool__(self):
        """Returns whether the configuration has any valuable configuration stored in it."""
        if self.message_id:
            return True
        
        return False
    
    
    def __eq__(self, other):
        """Returns whether the two configurations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # fail_fallback
        if self.fail_fallback != other.fail_fallback:
            return False
        
        # message_id
        if self.message_id != other.message_id:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # channel_id
        channel_id = self.channel_id
        if channel_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' channel_id = ')
            repr_parts.append(repr(channel_id))
        
        # message_id
        message_id = self.message_id
        if message_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
    
            repr_parts.append(' message_id = ')
            repr_parts.append(repr(message_id))
        
        # fail_fallback
        fail_fallback = self.fail_fallback
        if fail_fallback:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(', fail_fallback = ')
            repr_parts.append(repr(fail_fallback))
        
        # type
        message_reference_type = self.type
        if message_reference_type is not MessageReferenceType.reply:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
    
            repr_parts.append(' type = ')
            repr_parts.append(message_reference_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(message_reference_type.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the configuration's hash value."""
        hash_value = 0
        
        # channel_id
        hash_value ^= self.channel_id
        
        # fail_fallback
        hash_value ^= self.fail_fallback
        
        # message_id
        hash_value ^= self.message_id
        
        # type
        hash_value ^= self.type
        
        return hash_value
    
    
    def __or__(self, other):
        """Merges the two configurations"""
        if type(self) is not type(other):
            return NotImplemented
        
        # channel_id
        channel_id = other.channel_id
        if not channel_id:
            channel_id = self.channel_id
        
        # fail_fallback
        fail_fallback = other.fail_fallback
        if not fail_fallback:
            fail_fallback = self.fail_fallback
        
        # message_id
        message_id = other.message_id
        if not message_id:
            message_id = self.message_id
        
        # message_reference_type
        message_reference_type = other.type
        if message_reference_type is MessageReferenceType.reply:
            message_reference_type = self.type
        
        # Construct
        new = object.__new__(type(self))
        new.channel_id = channel_id
        new.fail_fallback = fail_fallback
        new.message_id = message_id
        new.type = message_reference_type
        return new



__all__ = ()

from .client_base import DiscordGatewayClientBase
from .client_shard import DiscordGatewayClientShard
from .client_sharder import DiscordGatewayClientSharder


def create_gateway(client):
    """
    Creates a gateway for the given count with the given shard count.
    
    Parameters
    ----------
    client : ``Client``
        The client to create the gateway for.
    
    Returns
    -------
    gateway : ``DiscordGatewayClientBase``
    """
    shard_count = client.shard_count
    if shard_count <= 1:
        gateway = DiscordGatewayClientShard(client, 0)
    else:
        gateway = DiscordGatewayClientSharder(client, shard_count, None)
    
    return gateway


def reshard_gateway(client):
    """
    Reshards the gateway to a new shard count. Trying to keep the existing instances as possible.
    
    Parameters
    ----------
    client : ``Client``
        The client to reshard its gateway of.
    
    Returns
    -------
    gateway : ``DiscordGatewayClientBase``
    """
    gateway = client.gateway
    shard_count = client.shard_count
    
    # Using shard gateway?
    if isinstance(gateway, DiscordGatewayClientShard):
        if shard_count <= 1:
            return gateway
        
        return DiscordGatewayClientSharder(client, shard_count, None)
    
    # Using sharder gateway?
    if isinstance(gateway, DiscordGatewayClientSharder):
        if shard_count <= 1:
            return gateway.get_gateway(0)
        
        return DiscordGatewayClientSharder(client, shard_count, gateway.gateways)
    
    # Default or unknown gateway type? Should not happen.
    return create_gateway(client) 



__all__ = ('EmbeddedActivityConfiguration',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_age_gated, parse_client_platform_configurations, parse_content_security_policy_exceptions_exist,
    parse_default_orientation_lock_state, parse_default_tablet_orientation_lock_state, parse_position,
    parse_preview_video_asset_id, put_age_gated, put_client_platform_configurations,
    put_content_security_policy_exceptions_exist, put_default_orientation_lock_state,
    put_default_tablet_orientation_lock_state, put_position, put_preview_video_asset_id,
    validate_age_gated, validate_client_platform_configurations, validate_content_security_policy_exceptions_exist,
    validate_default_orientation_lock_state, validate_default_tablet_orientation_lock_state, validate_position,
    validate_preview_video_asset_id
)
from .preinstanced import OrientationLockState


class EmbeddedActivityConfiguration(RichAttributeErrorBaseType):
    """
    Represents an embedded activity's configuration.
    
    Attributes
    ----------
    age_gated : `bool`
        Whether an age gate warning should show up for the using before using the activity.
    client_platform_configurations : `None | dict<PlatformType, ClientPlatformConfiguration>`
        The embedded activity's configuration for each platform.
    content_security_policy_exceptions_exist : `bool`
        Whether the activity has content security exceptions, this includes that the activity's developers may see
        your ip.
    default_orientation_lock_state : ``OrientationLockState``
        The activity's default orientation lock state.
    default_tablet_orientation_lock_state : ``OrientationLockState``
        The activity's default orientation lock state for tablets.
    position : `int`
        The activity's position in the embedded activity listing.
    preview_video_asset_id : `int`
        The activity's preview asset's identifier.
    """
    __slots__ = (
        'age_gated', 'client_platform_configurations', 'content_security_policy_exceptions_exist',
        'default_orientation_lock_state', 'default_tablet_orientation_lock_state', 'position', 'preview_video_asset_id'
    )
    
    def __new__(
        cls,
        *,
        age_gated = ...,
        client_platform_configurations = ...,
        content_security_policy_exceptions_exist = ...,
        default_orientation_lock_state = ...,
        default_tablet_orientation_lock_state = ...,
        position = ...,
        preview_video_asset_id = ...,
    ):
        """
        Creates a new embedded activity's configuration.
        
        Parameters
        ----------
        age_gated : `bool`, Optional (Keyword only)
            Whether an age gate warning should show up for the using before using the activity.
        client_platform_configurations : `None | dict<PlatformType | int, ClientPlatformConfiguration>` \
                , Optional (Keyword only)
            The embedded activity's configuration for each platform.
        content_security_policy_exceptions_exist : `bool`, Optional (Keyword only)
            Whether the activity has content security exceptions, this includes that the activity's developers may see
            your ip.
        default_orientation_lock_state : `OrientationLockState | int`, Optional (Keyword only)
            The activity's default orientation lock state.
        default_tablet_orientation_lock_state : `OrientationLockState | int`, Optional (Keyword only)
            The activity's default orientation lock state for tablets.
        position : `int`, Optional (Keyword only)
            The activity's position in the embedded activity listing.
        preview_video_asset_id : `int`, Optional (Keyword only)
            The activity's preview asset's identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # age_gated
        if age_gated is ...:
            age_gated = False
        else:
            age_gated = validate_age_gated(age_gated)
        
        # client_platform_configurations
        if client_platform_configurations is ...:
            client_platform_configurations = None
        else:
            client_platform_configurations = validate_client_platform_configurations(client_platform_configurations)
        
        # content_security_policy_exceptions_exist
        if content_security_policy_exceptions_exist is ...:
            content_security_policy_exceptions_exist = False
        else:
            content_security_policy_exceptions_exist = validate_content_security_policy_exceptions_exist(
                content_security_policy_exceptions_exist
            )
        
        # default_orientation_lock_state
        if default_orientation_lock_state is ...:
            default_orientation_lock_state = OrientationLockState.none
        else:
            default_orientation_lock_state = validate_default_orientation_lock_state(default_orientation_lock_state)
        
        # default_tablet_orientation_lock_state
        if default_tablet_orientation_lock_state is ...:
            default_tablet_orientation_lock_state = OrientationLockState.none
        else:
            default_tablet_orientation_lock_state = validate_default_tablet_orientation_lock_state(
                default_tablet_orientation_lock_state
            )
        
        # position
        if position is ...:
            position = 0
        else:
            position = validate_position(position)
        
        # preview_video_asset_id
        if preview_video_asset_id is ...:
            preview_video_asset_id = 0
        else:
            preview_video_asset_id = validate_preview_video_asset_id(preview_video_asset_id)
        
        # Construct
        self = object.__new__(cls)
        self.age_gated = age_gated
        self.client_platform_configurations = client_platform_configurations
        self.content_security_policy_exceptions_exist = content_security_policy_exceptions_exist
        self.default_orientation_lock_state = default_orientation_lock_state
        self.default_tablet_orientation_lock_state = default_tablet_orientation_lock_state
        self.position = position
        self.preview_video_asset_id = preview_video_asset_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a embedded activity configuration from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Client platform configuration data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.age_gated = parse_age_gated(data)
        self.client_platform_configurations = parse_client_platform_configurations(data)
        self.content_security_policy_exceptions_exist = parse_content_security_policy_exceptions_exist(data)
        self.default_orientation_lock_state = parse_default_orientation_lock_state(data)
        self.default_tablet_orientation_lock_state = parse_default_tablet_orientation_lock_state(data)
        self.position = parse_position(data)
        self.preview_video_asset_id = parse_preview_video_asset_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the embedded activity configuration.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_age_gated(self.age_gated, data, defaults)
        put_client_platform_configurations(self.client_platform_configurations, data, defaults)
        put_content_security_policy_exceptions_exist(self.content_security_policy_exceptions_exist, data, defaults)
        put_default_orientation_lock_state(self.default_orientation_lock_state, data, defaults)
        put_default_tablet_orientation_lock_state(self.default_tablet_orientation_lock_state, data, defaults)
        put_position(self.position, data, defaults)
        put_preview_video_asset_id(self.preview_video_asset_id, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the embedded activity configuration's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        field_added = False
        
        # age_gated
        age_gated = self.age_gated
        if age_gated:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' age_gated = ')
            repr_parts.append(repr(age_gated))
        
        # client_platform_configurations
        client_platform_configurations = self.client_platform_configurations
        if client_platform_configurations is not None:
            items = sorted(client_platform_configurations.items())
            length = len(items)
            
            # length should not be 0
            if length:
                if field_added:
                    repr_parts.append(',')
                else:
                    field_added = True
                
                repr_parts.append(' client_platform_configurations = {')
                
                index = 0
                while True:
                    platform, client_configuration = items[index]
                    
                    repr_parts.append(platform.name)
                    repr_parts.append(' ~ ')
                    repr_parts.append(repr(platform.value))
                    repr_parts.append(': ')
                    repr_parts.append(repr(client_configuration))
                    
                    index += 1
                    if index == length:
                        break
                    
                    repr_parts.append(', ')
                    continue
        
        # content_security_policy_exceptions_exist
        content_security_policy_exceptions_exist = self.content_security_policy_exceptions_exist
        if content_security_policy_exceptions_exist:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' content_security_policy_exceptions_exist = ')
            repr_parts.append(repr(content_security_policy_exceptions_exist))
        
        # default_orientation_lock_state
        default_orientation_lock_state = self.default_orientation_lock_state
        if default_orientation_lock_state is not OrientationLockState.none:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' default_orientation_lock_state = ')
            repr_parts.append(default_orientation_lock_state.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(default_orientation_lock_state.value))
        
        # default_tablet_orientation_lock_state
        default_tablet_orientation_lock_state = self.default_tablet_orientation_lock_state
        if default_tablet_orientation_lock_state is not OrientationLockState.none:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' default_tablet_orientation_lock_state = ')
            repr_parts.append(default_tablet_orientation_lock_state.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(default_tablet_orientation_lock_state.value))
        
        # position
        position = self.position
        if position:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' position = ')
            repr_parts.append(repr(position))
        
        # preview_video_asset_id
        preview_video_asset_id = self.preview_video_asset_id
        if preview_video_asset_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' preview_video_asset_id = ')
            repr_parts.append(repr(preview_video_asset_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activity configurations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # age_gated
        if self.age_gated != other.age_gated:
            return False
        
        # client_platform_configurations
        if self.client_platform_configurations != other.client_platform_configurations:
            return False
        
        # content_security_policy_exceptions_exist
        if self.content_security_policy_exceptions_exist != other.content_security_policy_exceptions_exist:
            return False
        
        # default_orientation_lock_state
        if self.default_orientation_lock_state is not other.default_orientation_lock_state:
            return False
        
        # default_tablet_orientation_lock_state
        if self.default_tablet_orientation_lock_state is not other.default_tablet_orientation_lock_state:
            return False
        
        # position
        if self.position != other.position:
            return False
        
        # preview_video_asset_id
        if self.preview_video_asset_id != other.preview_video_asset_id:
            return False
        
        return True
    

    def __hash__(self):
        """Returns the embedded activity configuration's hash value."""
        hash_value = 0
        
        # age_gated
        hash_value ^= self.age_gated
        
        # client_platform_configurations
        client_platform_configurations = self.client_platform_configurations
        if (client_platform_configurations is not None):
            for platform, client_configuration in client_platform_configurations.items():
                hash_value ^= hash(platform) & hash(client_configuration)
        
        # content_security_policy_exceptions_exist
        hash_value ^= self.content_security_policy_exceptions_exist << 4
        
        # default_orientation_lock_state
        hash_value ^= hash(self.default_orientation_lock_state) << 8
        
        # default_tablet_orientation_lock_state
        hash_value ^= hash(self.default_tablet_orientation_lock_state) << 12
        
        # position
        hash_value ^= self.position << 16
        
        # preview_video_asset_id
        hash_value ^= self.preview_video_asset_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the embedded activity configuration.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.age_gated = self.age_gated
        client_platform_configurations = self.client_platform_configurations
        if (client_platform_configurations is not None):
            client_platform_configurations = {
                platform: client_configuration.copy()
                for platform, client_configuration in client_platform_configurations.items()
            }
        new.client_platform_configurations = client_platform_configurations
        new.content_security_policy_exceptions_exist = self.content_security_policy_exceptions_exist
        new.default_orientation_lock_state = self.default_orientation_lock_state
        new.default_tablet_orientation_lock_state = self.default_tablet_orientation_lock_state
        new.position = self.position
        new.preview_video_asset_id = self.preview_video_asset_id
        return new
    
    
    def copy_with(
        self,
        *,
        age_gated = ...,
        client_platform_configurations = ...,
        content_security_policy_exceptions_exist = ...,
        default_orientation_lock_state = ...,
        default_tablet_orientation_lock_state = ...,
        position = ...,
        preview_video_asset_id = ...,
    ):
        """
        Copies the embedded activity configuration with the given fields.
        
        Parameters
        ----------
        age_gated : `bool`, Optional (Keyword only)
            Whether an age gate warning should show up for the using before using the activity.
        client_platform_configurations : `None | dict<PlatformType, ClientPlatformConfiguration>` \
                , Optional (Keyword only)
            The embedded activity's configuration for each platform.
        content_security_policy_exceptions_exist : `bool`, Optional (Keyword only)
            Whether the activity has content security exceptions, this includes that the activity's developers may see
            your ip.
        default_orientation_lock_state : `OrientationLockState | int`, Optional (Keyword only)
            The activity's default orientation lock state.
        default_tablet_orientation_lock_state : `OrientationLockState | int`, Optional (Keyword only)
            The activity's default orientation lock state for tablets.
        position : `int`, Optional (Keyword only)
            The activity's position in the embedded activity listing.
        preview_video_asset_id : `int`, Optional (Keyword only)
            The activity's preview asset's identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # age_gated
        if age_gated is ...:
            age_gated = self.age_gated
        else:
            age_gated = validate_age_gated(age_gated)
        
        # client_platform_configurations
        if client_platform_configurations is ...:
            client_platform_configurations = self.client_platform_configurations
            if (client_platform_configurations is not None):
                client_platform_configurations = {
                    platform: client_configuration.copy()
                    for platform, client_configuration in client_platform_configurations.items()
                }
        else:
            client_platform_configurations = validate_client_platform_configurations(client_platform_configurations)
        
        # content_security_policy_exceptions_exist
        if content_security_policy_exceptions_exist is ...:
            content_security_policy_exceptions_exist = self.content_security_policy_exceptions_exist
        else:
            content_security_policy_exceptions_exist = validate_content_security_policy_exceptions_exist(
                content_security_policy_exceptions_exist
            )
        
        # default_orientation_lock_state
        if default_orientation_lock_state is ...:
            default_orientation_lock_state = self.default_orientation_lock_state
        else:
            default_orientation_lock_state = validate_default_orientation_lock_state(default_orientation_lock_state)
        
        # default_tablet_orientation_lock_state
        if default_tablet_orientation_lock_state is ...:
            default_tablet_orientation_lock_state = self.default_tablet_orientation_lock_state
        else:
            default_tablet_orientation_lock_state = validate_default_tablet_orientation_lock_state(
                default_tablet_orientation_lock_state
            )
        
        # position
        if position is ...:
            position = self.position
        else:
            position = validate_position(position)
        
        # preview_video_asset_id
        if preview_video_asset_id is ...:
            preview_video_asset_id = self.preview_video_asset_id
        else:
            preview_video_asset_id = validate_preview_video_asset_id(preview_video_asset_id)
        
        # Construct
        new = object.__new__(type(self))
        new.age_gated = age_gated
        new.client_platform_configurations = client_platform_configurations
        new.content_security_policy_exceptions_exist = content_security_policy_exceptions_exist
        new.default_orientation_lock_state = default_orientation_lock_state
        new.default_tablet_orientation_lock_state = default_tablet_orientation_lock_state
        new.position = position
        new.preview_video_asset_id = preview_video_asset_id
        return new
    
    
    def iter_supported_platforms(self):
        """
        Iterates over the supported platforms of the activity.
        
        This method is a generator.
        
        Yields
        ------
        platform : ``PlatformType`` 
        """
        client_platform_configurations = self.client_platform_configurations
        if (client_platform_configurations is not None):
            yield from client_platform_configurations.keys()
    
    
    def get_client_platform_configuration(self, platform):
        """
        Gets the client configuration for the given platform.
        
        Parameters
        ----------
        platform : `PlatformType | str`
            The platform to get the configuration for.
        
        Returns
        -------
        configuration : `None | ClientPlatformConfiguration`
        """
        client_platform_configurations = self.client_platform_configurations
        if (client_platform_configurations is not None):
            return client_platform_configurations.get(platform, None)
